@@ -16,6 +16,8 @@ G_BEGIN_DECLS
 /* enumerations from "ev-document-model.h" */
 GType ev_sizing_mode_get_type (void) G_GNUC_CONST;
 #define EV_TYPE_SIZING_MODE (ev_sizing_mode_get_type ())
+GType ev_page_layout_get_type (void) G_GNUC_CONST;
+#define EV_TYPE_PAGE_LAYOUT (ev_page_layout_get_type ())
 
 /* enumerations from "ev-job-scheduler.h" */
 GType ev_job_priority_get_type (void) G_GNUC_CONST;
